# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_mixing_elbow_settings_api_beta:

Fluent setup and solution using settings objects
------------------------------------------------
This example sets up and solves a three-dimensional turbulent fluid flow
and heat transfer problem in a mixing elbow, which is common in piping
systems in power plants and process industries. Predicting the flow field
and temperature field in the area of the mixing region is important to
designing the junction properly.

This example uses settings objects.

**Problem description**

A cold fluid at 20 deg C flows into the pipe through a large inlet. It then mixes
with a warmer fluid at 40 deg C that enters through a smaller inlet located at
the elbow. The pipe dimensions are in inches, and the fluid properties and
boundary conditions are given in SI units. Because the Reynolds number for the
flow at the larger inlet is ``50, 800``, a turbulent flow model is required.
"""

###############################################################################
# Perform required imports
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Perform required imports, which includes downloading and importing
# the geometry file.

# sphinx_gallery_thumbnail_path = '_static/mixing_elbow_settings.png'
import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

import_file_name = examples.download_file(
    "mixing_elbow.msh.h5", "pyfluent/mixing_elbow"
)

###############################################################################
# Launch Fluent
# ~~~~~~~~~~~~~
# Launch Fluent as a service in solver mode with double precision running on
# two processors and print Fluent version.

solver = pyfluent.launch_fluent(
    precision="double",
    processor_count=2,
    mode="solver",
)
print(solver.get_fluent_version())

###############################################################################
# Import mesh and perform mesh check
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Import the mesh and perform a mesh check, which lists the minimum and maximum
# x, y, and z values from the mesh in the default SI units of meters. The mesh
# check also reports a number of other mesh features that are checked. Any errors
# in the mesh are reported. Ensure that the minimum volume is not negative because
# Fluent cannot begin a calculation when this is the case.

solver.file.read_case(file_name=import_file_name)
solver.mesh.check()

###############################################################################
# Enable heat transfer
# ~~~~~~~~~~~~~~~~~~~~
# Enable heat transfer by activating the energy equation.

solver.setup.models.energy.enabled = True

###############################################################################
# Create material
# ~~~~~~~~~~~~~~~
# Create a material named ``"water-liquid"``.

solver.setup.materials.database.copy_by_name(type="fluid", name="water-liquid")

###############################################################################
# Set up cell zone conditions
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Set up the cell zone conditions for the fluid zone (``elbow-fluid``). Set ``material``
# to ``"water-liquid"``.

solver.setup.cell_zone_conditions.fluid["elbow-fluid"].general.material = "water-liquid"

###############################################################################
# Set up boundary conditions for CFD analysis
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Set up the boundary conditions for the inlets, outlet, and walls for CFD
# analysis.

# cold inlet (cold-inlet), Setting: Value:
# Velocity Specification Method: Magnitude, Normal to Boundary
# Velocity Magnitude: 0.4 [m/s]
# Specification Method: Intensity and Hydraulic Diameter
# Turbulent Intensity: 5 [%]
# Hydraulic Diameter: 4 [inch]
# Temperature: 293.15 [K]
cold_inlet = solver.setup.boundary_conditions.velocity_inlet["cold-inlet"]

cold_inlet.momentum.velocity_magnitude.value = 0.4
cold_inlet.turbulence.turbulence_specification = "Intensity and Hydraulic Diameter"
cold_inlet.turbulence.turbulent_intensity = 0.05
cold_inlet.turbulence.hydraulic_diameter = "4 [in]"
cold_inlet.thermal.temperature.value = 293.15

# hot inlet (hot-inlet), Setting: Value:
# Velocity Specification Method: Magnitude, Normal to Boundary
# Velocity Magnitude: 1.2 [m/s]
# Specification Method: Intensity and Hydraulic Diameter
# Turbulent Intensity: 5 [%]
# Hydraulic Diameter: 1 [inch]
# Temperature: 313.15 [K]
hot_inlet = solver.setup.boundary_conditions.velocity_inlet["hot-inlet"]

hot_inlet.momentum.velocity_magnitude.value = 1.2
hot_inlet.turbulence.turbulence_specification = "Intensity and Hydraulic Diameter"
hot_inlet.turbulence.hydraulic_diameter = "1 [in]"
hot_inlet.thermal.temperature.value = 313.15

# pressure outlet (outlet), Setting: Value:
# Backflow Turbulent Intensity: 5 [%]
# Backflow Turbulent Viscosity Ratio: 4

solver.setup.boundary_conditions.pressure_outlet[
    "outlet"
].turbulence.turbulent_viscosity_ratio = 4

###############################################################################
# Initialize flow field
# ~~~~~~~~~~~~~~~~~~~~~
# Initialize the flow field using hybrid initialization.

solver.solution.initialization.hybrid_initialize()

###############################################################################
# Solve for 150 iterations
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Solve for 150 iterations.

solver.solution.run_calculation.iterate(iter_count=150)

###############################################################################
# Configure graphics picture export
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Since Fluent is being run without the GUI, we will need to export plots as
# picture files. Edit the picture settings to use a custom resolution so that
# the images are large enough.

graphics = solver.results.graphics
# use_window_resolution option not active inside containers or Ansys Lab environment
if graphics.picture.use_window_resolution.is_active():
    graphics.picture.use_window_resolution = False
graphics.picture.x_resolution = 1920
graphics.picture.y_resolution = 1440

###############################################################################
# Create velocity vectors
# ~~~~~~~~~~~~~~~~~~~~~~~
# Create and display velocity vectors on the ``symmetry-xyplane`` plane, then
# export the image for inspection.

graphics = solver.results.graphics

graphics.vector["velocity_vector_symmetry"] = {}
velocity_symmetry = solver.results.graphics.vector["velocity_vector_symmetry"]
velocity_symmetry.print_state()
velocity_symmetry.field = "velocity-magnitude"
velocity_symmetry.surfaces_list = [
    "symmetry-xyplane",
]
velocity_symmetry.options.scale = 4
velocity_symmetry.style = "arrow"
velocity_symmetry.display()

graphics.views.restore_view(view_name="front")
graphics.views.auto_scale()
graphics.picture.save_picture(file_name="velocity_vector_symmetry.png")

###############################################################################
# .. image:: /_static/mixing_elbow_016.png
#   :width: 500pt
#   :align: center

###############################################################################
# Compute mass flow rate
# ~~~~~~~~~~~~~~~~~~~~~~
# Compute the mass flow rate.
solver.solution.report_definitions.flux["mass_flow_rate"] = {}

mass_flow_rate = solver.solution.report_definitions.flux["mass_flow_rate"]
mass_flow_rate.boundaries.allowed_values()
mass_flow_rate.boundaries = [
    "cold-inlet",
    "hot-inlet",
    "outlet",
]
mass_flow_rate.print_state()
solver.solution.report_definitions.compute(report_defs=["mass_flow_rate"])

#########################################################################
# Close Fluent
# ~~~~~~~~~~~~
# Close Fluent.

solver.exit()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Interceptor classes to use with gRPC services."""

import builtins
import logging
import os
from typing import Any

from google.protobuf.json_format import MessageToDict
from google.protobuf.message import DecodeError, Message
import grpc

from ansys.fluent.core.services.batch_ops import BatchOps

network_logger: logging.Logger = logging.getLogger("pyfluent.networking")
log_bytes_limit: int = int(os.getenv("PYFLUENT_GRPC_LOG_BYTES_LIMIT", 1000))
truncate_len: int = log_bytes_limit // 5


def _upper_snake_case_to_camel_case(name: str) -> str:
    return "".join([word.capitalize() for word in name.split("_") if word])


def _truncate_grpc_str(message: Message) -> str:
    message_bytes = message.ByteSize()
    message_str = str(MessageToDict(message))
    if not log_bytes_limit or message_bytes < log_bytes_limit:
        return message_str
    else:
        network_logger.debug(
            f"GRPC_TRACE: message partially hidden, {message_bytes} bytes > "
            f"{log_bytes_limit} bytes limit. To see the full message, set PYFLUENT_GRPC_LOG_BYTES_LIMIT to 0."
        )
        return f"{message_str[:truncate_len]} < ... > {message_str[-truncate_len:]}"


class TracingInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor class to trace gRPC calls."""

    def __init__(self) -> None:
        """__init__ method of TracingInterceptor class."""
        super().__init__()

    def _intercept_call(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        network_logger.debug(
            f"GRPC_TRACE: RPC = {client_call_details.method}, request = {_truncate_grpc_str(request)}"
        )
        response = continuation(client_call_details, request)
        if not response.exception():
            # call _truncate_grpc_str early to get the size warning even when hiding secrets
            response_str = _truncate_grpc_str(response.result())
            if os.getenv("PYFLUENT_HIDE_LOG_SECRETS") != "1":
                network_logger.debug(f"GRPC_TRACE: response = {response_str}")
        return response

    def intercept_unary_unary(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        """Intercept unary-unary call for tracing."""
        return self._intercept_call(continuation, client_call_details, request)


class ErrorStateInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor class to check Fluent server error state before gRPC calls are
    made."""

    def __init__(self, fluent_error_state) -> None:
        """__init__ method of ErrorStateInterceptor class."""
        super().__init__()
        self._fluent_error_state = fluent_error_state

    def _intercept_call(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        if self._fluent_error_state.name == "fatal":
            details = self._fluent_error_state.details
            raise RuntimeError(
                f"Fatal error identified on the Fluent server: {details}."
            )
        return continuation(client_call_details, request)

    def intercept_unary_unary(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        """Intercept unary-unary call for error state checking."""
        return self._intercept_call(continuation, client_call_details, request)


class GrpcErrorInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor class to check Fluent server error state before gRPC calls are
    made."""

    def _intercept_call(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        response = continuation(client_call_details, request)
        if response.exception() is not None and response.code() != grpc.StatusCode.OK:
            new_ex_cls = RuntimeError
            try:
                from google.rpc import error_details_pb2
                from grpc_status import rpc_status

                status = rpc_status.from_call(response)
                if status:
                    for detail in status.details:
                        if detail.Is(error_details_pb2.ErrorInfo.DESCRIPTOR):
                            info = error_details_pb2.ErrorInfo()
                            detail.Unpack(info)
                            if info.domain == "Python":
                                reason = info.reason
                                ex_cls_name = _upper_snake_case_to_camel_case(reason)
                                if hasattr(builtins, ex_cls_name):
                                    cls = getattr(builtins, ex_cls_name)
                                    if issubclass(cls, Exception):
                                        new_ex_cls = cls
                                        break
            except DecodeError:
                pass
            ex = response.exception()
            new_ex = new_ex_cls(
                ex.details() if isinstance(ex, grpc.RpcError) else str(ex)
            )
            new_ex.__context__ = ex
            raise new_ex from None
        return response

    def intercept_unary_unary(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        """Intercept unary-unary call for error state checking."""
        return self._intercept_call(continuation, client_call_details, request)


class BatchedFuture(grpc.Future):
    """Class implementing gRPC.Future interface.

    An instance of BatchedFuture is returned if the gRPC method is queued to be executed
    in batch later.
    """

    def __init__(self, result_cls) -> None:
        """__init__ method of BatchedFuture class."""
        self._result_cls = result_cls

    def cancel(self) -> bool:
        """Attempts to cancel the computation."""
        return False

    def cancelled(self) -> bool:
        """Describes whether the computation was cancelled."""
        return False

    def running(self) -> bool:
        """Describes whether the computation is taking place."""
        return False

    def done(self) -> bool:
        """Describes whether the computation has taken place."""
        return True

    def result(self, timeout=None) -> Any:
        """Returns the result of the computation or raises its exception."""
        return self._result_cls()

    def exception(self, timeout=None) -> None:
        """Return the exception raised by the computation."""
        return None

    def traceback(self, timeout=None) -> None:
        """Access the traceback of the exception raised by the computation."""
        return None

    def add_done_callback(self, fn) -> None:
        """Adds a function to be called at completion of the computation."""
        pass


class BatchInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor class to batch gRPC calls."""

    def __init__(self) -> None:
        """__init__ method of BatchInterceptor class."""
        super().__init__()

    def _intercept_call(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        batchOps = BatchOps.instance()
        if batchOps and batchOps.batching:
            qual_method = client_call_details.method
            package_and_service, method = qual_method.lstrip("/").split("/")
            package, service = package_and_service.rsplit(".", 1)
            op = batchOps.add_op(package, service, method, request)
            if op.queued:
                return BatchedFuture(op.response_cls)

        return continuation(client_call_details, request)

    def intercept_unary_unary(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        """Intercept unary-unary call for batch operation."""
        return self._intercept_call(continuation, client_call_details, request)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core import examples
from ansys.fluent.core.parametric import (
    LocalParametricStudy,
    convert_design_point_parameter_units,
)


@pytest.mark.standalone
def test_local_parametric_run():
    case_filepath = examples.download_file(
        "Static_Mixer_Parameters.cas.h5",
        "pyfluent/static_mixer",
        return_without_path=False,
    )

    local_study = LocalParametricStudy(case_filepath=case_filepath)

    for idx in range(4):
        design_point = local_study.add_design_point("dp_" + str(idx))
        design_point.input_parameters["inlet1_vel"] = float(2 + idx)

    local_study.run_in_fluent(num_servers=2)

    table = local_study.design_point_table

    assert len(table) == 5

    for point in table:
        ins = convert_design_point_parameter_units(point.input_parameters)
        outs = point.output_parameters
        new_inlet_velocity = ins["inlet1_vel"]
        new_outlet_velocity = outs["outlet-vel-avg-op"]
        assert new_inlet_velocity
        assert new_outlet_velocity



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os
from unittest import TestCase

import pytest
from util.solver import SettingsValDict as D
from util.solver import assign_settings_value_from_value_dict as assign_dict_val
from util.solver import get_name_info


@pytest.mark.fluent_version(">=24.1")
@pytest.mark.settings_only
@pytest.mark.codegen_required
def test_boundaries_elbow(mixing_elbow_settings_session):
    solver_session = mixing_elbow_settings_session
    solver_session.setup.models.energy.enabled = True

    cold_inlet = solver_session.setup.boundary_conditions.velocity_inlet["cold-inlet"]
    assert D(1) == cold_inlet.momentum.velocity()
    assign_dict_val(cold_inlet.momentum.velocity, 0.4)
    assert D(0.4) == cold_inlet.momentum.velocity()

    cold_inlet.turbulence.turbulent_specification = "Intensity and Hydraulic Diameter"
    cold_inlet.turbulence.turbulent_intensity = 0.05
    cold_inlet.turbulence.hydraulic_diameter = "4 [in]"
    assign_dict_val(cold_inlet.thermal.t, 293.15)

    assert {
        "name": "cold-inlet",
        "momentum": {
            "initial_gauge_pressure": {"option": "value", "value": 0},
            "reference_frame": "Absolute",
            "velocity": {"option": "value", "value": 0.4},
            "velocity_specification_method": "Magnitude, Normal to Boundary",
        },
        "turbulence": {
            "turbulent_specification": "Intensity and Hydraulic Diameter",
            "turbulent_intensity": 0.05,
            "hydraulic_diameter": "4 [in]",
        },
        "thermal": {"t": {"option": "value", "value": 293.15}},
    } == cold_inlet()

    hot_inlet = solver_session.setup.boundary_conditions.velocity_inlet["hot-inlet"]
    assign_dict_val(hot_inlet.momentum.velocity, 1.2)
    hot_inlet.turbulence.turbulent_specification = "Intensity and Hydraulic Diameter"
    hot_inlet.turbulence.hydraulic_diameter = "1 [in]"
    assign_dict_val(hot_inlet.thermal.t, 313.15)

    assert {
        "name": "hot-inlet",
        "momentum": {
            "initial_gauge_pressure": {"option": "value", "value": 0},
            "reference_frame": "Absolute",
            "velocity": {"option": "value", "value": 1.2},
            "velocity_specification_method": "Magnitude, Normal to Boundary",
        },
        "turbulence": {
            "turbulent_specification": "Intensity and Hydraulic Diameter",
            "turbulent_intensity": 0.05,
            "hydraulic_diameter": "1 [in]",
        },
        "thermal": {"t": {"option": "value", "value": 313.15}},
    } == hot_inlet()

    solver_session.setup.boundary_conditions.pressure_outlet[
        "outlet"
    ].turbulence.turbulent_viscosity_ratio = 4
    assert (
        solver_session.setup.boundary_conditions.pressure_outlet[
            "outlet"
        ].turbulence.turbulent_viscosity_ratio()
        == 4
    )


@pytest.mark.settings_only
@pytest.mark.fluent_version("latest")
def test_boundaries_periodic(periodic_rot_settings_session):
    solver_session = periodic_rot_settings_session
    print(__file__)
    _THIS_DIR = os.path.dirname(__file__)
    _DATA_FILE = os.path.join(_THIS_DIR, "boundaries_periodic_expDict")
    boundary_exp = json.load(open(_DATA_FILE))
    boundary_test = dict()
    boundary_tested = dict()
    for name, boundary in solver_session.setup.boundary_conditions.items():
        boundary_test[name] = boundary()
    boundary_tested["val_1"] = boundary_test

    TestCase().assertDictEqual(boundary_tested["val_1"], boundary_exp["val_1"])

    boundary_test = dict()
    for (
        boundary_type
    ) in solver_session.setup.boundary_conditions.get_active_child_names():
        if boundary_type in ["non_reflecting_bc", "perforated_wall", "settings"]:
            continue
        for name, boundary in getattr(
            solver_session.setup.boundary_conditions, boundary_type
        ).items():
            boundary_test[name] = boundary()
    boundary_tested["val_2"] = boundary_test
    TestCase().assertDictEqual(boundary_tested["val_2"], boundary_exp["val_2"])

    boundaries_check = ["inlet", "outlet", "pipe_2_wall"]
    selected_bou_test = get_name_info(boundary_tested["val_1"], boundaries_check)
    selected_bou_exp = get_name_info(boundary_exp["val_1"], boundaries_check)
    TestCase().assertDictEqual(selected_bou_test, selected_bou_exp)
    solver_session.setup.boundary_conditions.wall["pipe_2_wall"].rename("pipe2_wall")
    solver_session.setup.boundary_conditions.pressure_outlet["outlet"].rename("out")
    solver_session.setup.boundary_conditions.velocity_inlet[
        "inlet"
    ].momentum.velocity = 5.0
    solver_session.setup.boundary_conditions["inlet"].momentum.velocity = 10.0
    boundaries_check = ["inlet", "out", "pipe2_wall"]
    boundary_test = dict()
    for name, boundary in solver_session.setup.boundary_conditions.items():
        boundary_test[name] = boundary()
    boundary_tested["val_3"] = boundary_test
    TestCase().assertDictEqual(boundary_tested["val_3"], boundary_exp["val_3"])

    selected_bou_test = get_name_info(boundary_tested["val_3"], boundaries_check)
    selected_bou_exp = get_name_info(boundary_exp["val_3"], boundaries_check)
    TestCase().assertDictEqual(selected_bou_test, selected_bou_exp)
    with open("boundaries_periodic_outDict.py", "a") as f:
        json.dump(boundary_tested, f, sort_keys=True, indent=4)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides a ConversionStrategy for mapping VariableDescriptor to variable names used in Fluent expressions.
"""

from ansys.units.variable_descriptor import (
    MappingConversionStrategy,
    VariableCatalog,
)


class FluentExprNamingStrategy(MappingConversionStrategy):
    """This strategy handles conversion of selected VariableCatalog into Fluent's
    server-side expression variable naming conventions.
    """

    _c = VariableCatalog

    _mapping = {
        # pressure
        _c.PRESSURE: "StaticPressure",
        _c.STATIC_PRESSURE: "StaticPressure",
        _c.ABSOLUTE_PRESSURE: "AbsolutePressure",
        _c.DYNAMIC_PRESSURE: "DynamicPressure",
        _c.TOTAL_PRESSURE: "TotalPressure",
        _c.PRESSURE_COEFFICIENT: "PressureCoefficient",
        # velocity
        _c.AXIAL_VELOCITY: "AxialVelocity",
        _c.CONVECTIVE_COURANT_NUMBER: "ElementConvectionCourantNumber",
        _c.CELL_REYNOLDS_NUMBER: "ElementReynoldsNumber",
        _c.fluent.HELICITY: "Helicity",
        _c.fluent.LAMBDA_2_CRITERION: "Lambda2Criterion",
        _c.MESH_VELOCITY: "MeshVelocity",
        _c.MESH_VELOCITY_X: "MeshVelocity.x",
        _c.MESH_VELOCITY_Y: "MeshVelocity.y",
        _c.MESH_VELOCITY_Z: "MeshVelocity.z",
        _c.MESH_VELOCITY_MAGNITUDE: "MeshVelocityMagnitude",
        _c.NORMALIZED_Q_CRITERION: "QCriterionNormalized",
        _c.Q_CRITERION: "QCriterionRaw",
        _c.RADIAL_VELOCITY: "RadialVelocity",
        _c.TANGENTIAL_VELOCITY: "TangentialVelocity",
        _c.VELOCITY: "Velocity",
        _c.VELOCITY_X: "Velocity.x",
        _c.VELOCITY_Y: "Velocity.y",
        _c.VELOCITY_Z: "Velocity.z",
        _c.VELOCITY_MAGNITUDE: "VelocityMagnitude",
        _c.fluent.VELOCITY_ANGLE: "VelocityAngle",
        _c.VORTICITY: "Vorticity",
        _c.VORTICITY_X: "Vorticity.x",
        _c.VORTICITY_Y: "Vorticity.y",
        _c.VORTICITY_Z: "Vorticity.z",
        _c.VORTICITY_MAGNITUDE: "VorticityMagnitude",
        # density
        _c.DENSITY: "Density",
        _c.fluent.DENSITY_ALL: "DensityAll",
        # "properties"
        _c.DYNAMIC_VISCOSITY: "DynamicViscosity",
        _c.PRANDTL_NUMBER: "PrandtlNumber",
        _c.SPECIFIC_HEAT_CAPACITY: "SpecificHeatCapacity",
        _c.THERMAL_CONDUCTIVITY: "ThermalConductivity",
        # turbulence
        _c.EFFECTIVE_PRANDTL_NUMBER: "EffectivePrandtlNumber",
        _c.EFFECTIVE_THERMAL_CONDUCTIVITY: "EffectiveThermalConductivity",
        _c.EFFECTIVE_VISCOSITY: "EffectiveViscosity",
        _c.PRODUCTION_OF_TURBULENT_KINETIC_ENERGY: "Productionofk",
        _c.SPECIFIC_DISSIPATION_RATE: "SpecificDissipationRateOmega",
        _c.TURBULENT_DISSIPATION_RATE: "TurbulenceDissipationRate",
        _c.TURBULENT_INTENSITY: "TurbulenceIntensity",
        _c.TURBULENT_VISCOSITY: "TurbulenceViscosity",
        _c.TURBULENT_VISCOSITY_RATIO: "TurbulenceViscosityRatio",
        _c.TURBULENT_KINETIC_ENERGY: "TurbulentKineticEnergyk",
        _c.TURBULENT_REYNOLDS_NUMBER: "TurbulentReynoldsNumberRe_y",
        _c.WALL_Y_PLUS: "WallYplus",
        _c.WALL_Y_STAR: "WallYstar",
        # wall fluxes
        _c.SURFACE_HEAT_TRANSFER_COEFFICIENT: "HeatTransferCoefficient",
        _c.SKIN_FRICTION_COEFFICIENT: "SkinFrictionCoefficient",
        _c.SURFACE_HEAT_FLUX: "SurfaceHeatFlux",
        _c.SURFACE_HEAT_TRANSFER_COEFFICIENT: "SurfaceHeatTransferCoefficient",
        _c.SURFACE_NUSSELT_NUMBER: "SurfaceNusseltNumber",
        _c.SURFACE_STANTON_NUMBER: "SurfaceStantonNumber",
        _c.WALL_ADJACENT_HEAT_TRANSFER_COEFFICIENT: "WallAdjacentHeatTransferCoef",
        _c.WALL_SHEAR_STRESS: "WallShearStressVector",
        _c.WALL_SHEAR_STRESS_X: "WallShearStressVector.x",
        _c.WALL_SHEAR_STRESS_Y: "WallShearStressVector.y",
        _c.WALL_SHEAR_STRESS_Z: "WallShearStressVector.z",
        _c.WALL_SHEAR_STRESS_MAGNITUDE: "WallShearStressVector.mag",
        _c.fluent.Y_PLUS_BASED_HEAT_TRANSFER_COEFFICIENT: "YplusBasedHeatTranCoef",
        # residuals
        _c.fluent.MASS_IMBALANCE: "MassImbalance",
        # derivatives
        _c.fluent.PRESSURE_HESSIAN_INDICATOR: "PressureHessianIndicator",
        _c.STRAIN_RATE: "StrainRate",
        _c.fluent.DVELOCITY_DX: "dVelocitydx",
        _c.fluent.DVELOCITY_DX_X: "dVelocitydx.x",
        _c.fluent.DVELOCITY_DX_Y: "dVelocitydx.y",
        _c.fluent.DVELOCITY_DX_Z: "dVelocitydx.z",
        _c.fluent.DVELOCITY_DX_MAGNITUDE: "dVelocitydx.mag",
        _c.fluent.DVELOCITY_DY: "dVelocitydy",
        _c.fluent.DVELOCITY_DY_X: "dVelocitydy.x",
        _c.fluent.DVELOCITY_DY_Y: "dVelocitydy.y",
        _c.fluent.DVELOCITY_DY_Z: "dVelocitydy.z",
        _c.fluent.DVELOCITY_DY_MAGNITUDE: "dVelocitydy.mag",
        _c.fluent.DVELOCITY_DZ: "dVelocitydz",
        _c.fluent.DVELOCITY_DZ_X: "dVelocitydz.x",
        _c.fluent.DVELOCITY_DZ_Y: "dVelocitydz.y",
        _c.fluent.DVELOCITY_DZ_Z: "dVelocitydz.z",
        _c.fluent.DVELOCITY_DZ_MAGNITUDE: "dVelocitydz.mag",
        # temperature
        _c.SPECIFIC_ENTHALPY: "SpecificEnthalpy",
        _c.SPECIFIC_ENTROPY: "SpecificEntropy",
        _c.SPECIFIC_INTERNAL_ENERGY: "SpecificInternalEnergy",
        _c.SPECIFIC_TOTAL_ENERGY: "SpecificTotalEnergy",
        _c.TEMPERATURE: "StaticTemperature",
        _c.SPECIFIC_TOTAL_ENTHALPY: "SpecificTotalEnthalpy",
        _c.fluent.TOTAL_ENTHALPY_DEVIATION: "TotalEnthalpyDeviation",
        _c.TOTAL_TEMPERATURE: "TotalTemperature",
        _c.WALL_ADJACENT_TEMPERATURE: "WallAdjacentTemperature",
        _c.WALL_TEMPERATURE: "WallTemperature",
        _c.WALL_TEMPERATURE_THIN: "WallTemperatureThin",
        _c.fluent.Y_PLUS_BASED_HEAT_TRANSFER_COEFFICIENT: "YplusBasedHeatTranRefTemperature",
        # mesh
        _c.mesh.ANISOTROPIC_ADAPTION_CELLS: "AnisotropicAdaptionCells",
        _c.mesh.BOUNDARY_CELL_DISTANCE: "BoundaryCellDistance",
        _c.mesh.BOUNDARY_LAYER_CELLS: "BoundaryLayerCells",
        _c.mesh.BOUNDARY_NORMAL_DISTANCE: "BoundaryNormalDistance",
        _c.mesh.BOUNDARY_VOLUME_DISTANCE: "BoundaryVolumeDistance",
        _c.mesh.CELL_EQUIANGLE_SKEW: "CellEquiangleSkew",
        _c.mesh.CELL_EQUIVOLUME_SKEW: "CellEquivolumeSkew",
        _c.mesh.CELL_PARENT_INDEX: "CellParentIndex",
        _c.mesh.CELL_REFINE_LEVEL: "CellRefineLevel",
        _c.mesh.CELL_VOLUME: "CellVolume",
        _c.mesh.CELL_VOLUME_CHANGE: "CellVolumeChange",
        _c.mesh.ELEMENT_ASPECT_RATIO: "ElementAspectRatio",
        _c.mesh.ELEMENT_WALL_DISTANCE: "ElementWallDistance",
        _c.mesh.FACE_AREA_MAGNITUDE: "FaceAreaMagnitude",
        _c.mesh.FACE_HANDEDNESS: "FaceHandedness",
        _c.mesh.INTERFACE_OVERLAP_FRACTION: "InterfaceOverlapFraction",
        _c.mesh.MARK_POOR_ELEMENTS: "MarkPoorElements",
        _c.POSITION: "Position",
        _c.POSITION_X: "Position.x",
        _c.POSITION_Y: "Position.y",
        _c.POSITION_Z: "Position.z",
        _c.POSITION_MAGNITUDE: "Position.mag",
        _c.mesh.SMOOTHED_CELL_REFINE_LEVEL: "SmoothedCellRefineLevel",
        _c.mesh.X_FACE_AREA: "XFaceArea",
        _c.mesh.Y_FACE_AREA: "YFaceArea",
        _c.mesh.Z_FACE_AREA: "ZFaceArea",
        _c.mesh.ACTIVE_CELL_PARTITION: "ActiveElementPartition",
        _c.mesh.CELL_ELEMENT_TYPE: "CellElementType",
        _c.mesh.CELL_ID: "CellId",
        _c.mesh.CELL_WEIGHT: "CellWeight",
        _c.mesh.CELL_ZONE_INDEX: "CellZoneIndex",
        _c.mesh.CELL_ZONE_TYPE: "CellZoneType",
        _c.mesh.PARTITION_NEIGHBOURS: "PartitionNeighbors",
        _c.mesh.STORED_CELL_PARTITIION: "StoredElementPartition",
    }



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module to get networking functionality."""

from concurrent import futures
import logging
import socket
from typing import Any
import urllib.request

import grpc
from grpc_health.v1 import health_pb2, health_pb2_grpc

network_logger = logging.getLogger("pyfluent.networking")


def get_free_port() -> int:
    """Identifies a free port to which a new socket connection can be established.

    Returns
    -------
    int
        port number
    """
    with socket.socket() as s:
        s.bind(("localhost", 0))
        free_port = s.getsockname()[1]
    return free_port


class _HealthServicer(health_pb2_grpc.HealthServicer):
    def Check(self, request, context: grpc.ServicerContext):
        """Check the status of service."""
        return health_pb2.HealthCheckResponse(
            status=health_pb2.HealthCheckResponse.ServingStatus.SERVING
        )


class _GrpcServer:
    def __init__(self, address: str):
        self._server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        self._server.add_insecure_port(address)
        health_pb2_grpc.add_HealthServicer_to_server(_HealthServicer(), self._server)
        self._server.start()

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        self._server.stop(None)


def find_remoting_ip() -> str:
    """Find an ip address at which a gRPC connection can be established by looping over
    getaddrinfo output.

    Returns
    -------
    str
        remoting ip address
    """
    from ansys.fluent.core import INFER_REMOTING_IP_TIMEOUT_PER_IP

    for addrinfo in socket.getaddrinfo(
        "localhost",
        0,
        family=socket.AF_INET,
        type=socket.SOCK_STREAM,
        flags=socket.AI_PASSIVE,
    ):
        ip = addrinfo[-1][0]
        port = get_free_port()
        address = f"{ip}:{port}"
        with _GrpcServer(address):
            with grpc.insecure_channel(address) as channel:
                stub = health_pb2_grpc.HealthStub(channel)
                try:
                    if (
                        stub.Check(
                            health_pb2.HealthCheckRequest(),
                            timeout=INFER_REMOTING_IP_TIMEOUT_PER_IP,
                        ).status
                        == health_pb2.HealthCheckResponse.ServingStatus.SERVING
                    ):
                        network_logger.debug(f"Can use {ip} as remoting ip")
                        return ip
                except Exception:
                    network_logger.debug(f"Cannot use {ip} as remoting ip")


def check_url_exists(url: str) -> bool:
    """Check if a URL exists.

    Parameters
    ----------
    url : str
        URL to check

    Returns
    -------
    bool
        True if the URL exists, False otherwise
    """
    try:
        with urllib.request.urlopen(url) as response:
            return response.status == 200
    except Exception:
        return False


def get_url_content(url: str) -> str:
    """Get the content of a URL.

    Parameters
    ----------
    url : str
        URL to get content from

    Returns
    -------
    str
        content of the URL
    """
    with urllib.request.urlopen(url) as response:
        return response.read()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing class encapsulating Fluent connection.

Expose aero capabilities.
"""

from typing import Any, Dict

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.services import SchemeEval
from ansys.fluent.core.services.datamodel_se import PyMenuGeneric
from ansys.fluent.core.session_solver import Solver


class SolverAero(Solver):
    """Encapsulates a Fluent server for Aero session connection.

    SolverAero(Session) holds the top-level objects for solver TUI, settings and aero
    datamodel objects calls.
    """

    def __init__(
        self,
        fluent_connection: FluentConnection,
        scheme_eval: SchemeEval,
        file_transfer_service: Any | None = None,
        start_transcript: bool = True,
        launcher_args: Dict[str, Any] | None = None,
    ):
        """SolverAero session.

        Parameters
        ----------
        fluent_connection (:ref:`ref_fluent_connection`):
            Encapsulates a Fluent connection.
        scheme_eval: SchemeEval
            Instance of ``SchemeEval`` to execute Fluent's scheme code on.
        file_transfer_service : Optional
            Service for uploading and downloading files.
        start_transcript : bool, optional
            Whether to start the Fluent transcript in the client.
            The default is ``True``, in which case the Fluent
            transcript can be subsequently started and stopped
            using method calls on the ``Session`` object.
        """
        super(SolverAero, self).__init__(
            fluent_connection=fluent_connection,
            scheme_eval=scheme_eval,
            file_transfer_service=file_transfer_service,
            start_transcript=start_transcript,
            launcher_args=launcher_args,
        )
        self._flserver_root = None
        self._fluent_version = None
        self._fluent_connection = fluent_connection
        # TODO: Update Aero DM
        scheme_eval.eval("(aero-load-addon)")

    def new_project(self, project_name: str):
        """Define a new project."""
        # TODO: Update Aero DM
        self.scheme.eval(f"""(prjapp-new-project-cb #f "{project_name}")""")

    def open_project(self, project_name: str):
        """Open a saved project."""
        # TODO: Update Aero DM
        self.scheme.eval(f"""(prjapp-project-open-project-cb #f "{project_name}")""")

    def new_simulation(self, case_file_name: str):
        """Add a new simulation by loading a case-file."""
        # TODO: Update Aero DM
        self.scheme.eval(
            f"""(gui-aero-project-add-workflow-cb #f "{case_file_name}" #f #f)"""
        )

    def open_simulation(self, simulation_file_name: str):
        """Open a saved simulation."""
        # TODO: Update Aero DM
        self.scheme.eval(
            f"""(aero-server-project-open-simulation "{simulation_file_name}")"""
        )

    @property
    def _flserver(self):
        """Root datamodel object."""
        return PyMenuGeneric(service=self._se_service, rules="flserver")

    @property
    def aero(self):
        """Instance of aero (Case.App) -> root datamodel object."""
        return self._flserver.Case.App

    def __dir__(self):
        return super(SolverAero, self).__dir__()



"""Provides a module for generating Fluent RST files."""

import importlib

from rstgen import _get_file_or_folder, generate

if __name__ == "__main__":
    try:
        meshing_tui = importlib.import_module(
            f"ansys.fluent.core.generated.meshing.{_get_file_or_folder(mode='meshing', is_datamodel=False)}"
        )
        generate(main_menu=meshing_tui.main_menu, mode="meshing", is_datamodel=False)
    except ModuleNotFoundError:
        pass
    try:
        solver_tui = importlib.import_module(
            f"ansys.fluent.core.generated.solver.{_get_file_or_folder(mode='solver', is_datamodel=False)}"
        )
        generate(main_menu=solver_tui.main_menu, mode="solver", is_datamodel=False)
    except ModuleNotFoundError:
        pass



"""Provides a module for generating Fluent datamodel RST files."""

import importlib

from rstgen import _get_file_or_folder, _write_datamodel_index_doc, generate


def generate_meshing_datamodels():
    """Generate meshing datamodel RST files."""
    meshing_datamodel_roots = []
    available_datamodels = []
    meshing_datamodels = [
        "meshing",
        "meshing_utilities",
        "part_management",
        "pm_file_management",
        "preferences",
        "workflow",
    ]
    for meshing_datamodel in meshing_datamodels:
        try:
            datamodel = importlib.import_module(
                f"ansys.fluent.core.generated.{_get_file_or_folder(mode='meshing', is_datamodel=True)}.{meshing_datamodel}"
            )
            if datamodel:
                meshing_datamodel_roots.append(datamodel.Root)
                available_datamodels.append(meshing_datamodel)
        except ModuleNotFoundError:
            pass
    _write_datamodel_index_doc(available_datamodels, "meshing")
    for root in meshing_datamodel_roots:
        generate(main_menu=root, mode="meshing", is_datamodel=True)


def generate_solver_datamodels():
    """Generate solver datamodel RST files."""
    solver_datamodel_roots = []
    available_datamodels = []
    solver_datamodels = ["flicing", "preferences", "solver_workflow", "workflow"]
    for solver_datamodel in solver_datamodels:
        try:
            datamodel = importlib.import_module(
                f"ansys.fluent.core.generated.{_get_file_or_folder(mode='solver', is_datamodel=True)}.{solver_datamodel}"
            )
            if datamodel:
                solver_datamodel_roots.append(datamodel.Root)
                available_datamodels.append(solver_datamodel)
        except ModuleNotFoundError:
            pass
    _write_datamodel_index_doc(available_datamodels, "solver")
    for root in solver_datamodel_roots:
        generate(main_menu=root, mode="solver", is_datamodel=True)


if __name__ == "__main__":
    generate_meshing_datamodels()
    generate_solver_datamodels()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Session utilities."""

from typing import Any, Dict

import ansys.fluent.core as pyfluent
from ansys.fluent.core.launcher.container_launcher import DockerLauncher
from ansys.fluent.core.launcher.launch_options import (
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    Precision,
    UIMode,
)
from ansys.fluent.core.launcher.pim_launcher import PIMLauncher
from ansys.fluent.core.launcher.standalone_launcher import StandaloneLauncher
from ansys.fluent.core.utils.fluent_version import FluentVersion


class SessionBase:
    """Base class for Fluent sessions.

    This class is not intended to be used directly. Instead, use
    the `from_connection`, `from_container`, `from_install`,
    or `from_pim` functions to create a session.
    """

    _session_mode = {
        "Meshing": FluentMode.MESHING,
        "PureMeshing": FluentMode.PURE_MESHING,
        "PrePost": FluentMode.PRE_POST,
        "Solver": FluentMode.SOLVER,
        "SolverAero": FluentMode.SOLVER_AERO,
        "SolverIcing": FluentMode.SOLVER_ICING,
    }

    @classmethod
    def from_install(
        cls,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        journal_file_names: None | str | list[str] = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        env: Dict[str, Any] = {},  # noqa: B006
        cleanup_on_exit: bool = True,
        dry_run: bool = False,
        start_transcript: bool = True,
        case_file_name: str | None = None,
        case_data_file_name: str | None = None,
        lightweight_mode: bool | None = None,
        py: bool | None = None,
        gpu: bool | None = None,
        cwd: str | None = None,
        fluent_path: str | None = None,
        topy: str | list | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
    ):
        """
        Launch a Fluent session in standalone mode.

        Parameters
        ----------
        ui_mode : UIMode
            Defines the user interface mode for Fluent. Options correspond to values in the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version : FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values are from the ``Dimension`` enum
            (``Dimension.TWO`` or ``Dimension.THREE``) or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values are from the ``Precision`` enum (``Precision.SINGLE``
            or ``Precision.DOUBLE``) or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        journal_file_names : str or list of str, optional
            Path(s) to a Fluent journal file(s) that Fluent will execute. Defaults to ``None``.
        start_timeout : int, optional
            Maximum time in seconds allowed for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        env : dict[str, str], optional
            A mapping for modifying environment variables in Fluent. Defaults to ``None``.
        cleanup_on_exit : bool, optional
            Determines whether to shut down the connected Fluent session when exiting PyFluent or calling
            the session's `exit()` method. Defaults to True.
        dry_run : bool, optional
            If True, does not launch Fluent but prints configuration information instead. The `call()` method
            returns a tuple containing the launch string and server info file name. Defaults to False.
        start_transcript : bool, optional
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        case_file_name : str, optional
            Name of the case file to read into the Fluent session. Defaults to None.
        case_data_file_name : str, optional
            Name of the case data file. If both case and data files are provided, they are read into the session.
        lightweight_mode : bool, optional
            If True, runs in lightweight mode where mesh settings are read into a background solver session,
            replacing it once complete. This parameter is only applicable when `case_file_name` is provided; defaults to False.
        py : bool, optional
            If True, runs Fluent in Python mode. Defaults to None.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        cwd : str, optional
            Working directory for the Fluent client.
        fluent_path: str, optional
            User-specified path for Fluent installation.
        topy :  bool or str, optional
            A flag indicating whether to write equivalent Python journals from provided journal files; can also specify
            a filename for the new Python journal.
        start_watchdog : bool, optional
            When `cleanup_on_exit` is True, defaults to True; an independent watchdog process ensures that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any
            Service for uploading/downloading files to/from the server.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        mode = cls._session_mode[cls.__name__]
        argvals = locals().copy()
        argvals.pop("cls", None)  # Remove the class reference from the arguments
        launcher = StandaloneLauncher(**argvals)
        return launcher()

    @classmethod
    def from_container(
        cls,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        container_dict: dict | None = None,
        dry_run: bool = False,
        cleanup_on_exit: bool = True,
        start_transcript: bool = True,
        py: bool | None = None,
        gpu: bool | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
    ):
        """
        Launch a Fluent session in container mode.

        Parameters
        ----------
        ui_mode : UIMode
            Defines the user interface mode for Fluent. Options correspond to values in the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version :  FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            any of ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values include ``Dimension.TWO``,
            ``Dimension.THREE``, or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values include ``Precision.SINGLE``,
            ``Precision.DOUBLE``, or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        start_timeout : int, optional
            Maximum allowable time in seconds for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        container_dict : dict, optional
            Configuration dictionary for launching Fluent inside a Docker container. See also
            :mod:`~ansys.fluent.core.launcher.fluent_container`.
        dry_run : bool, optional
            If True, does not launch Fluent but prints configuration information instead. If dry running a
            container start, this method will return the configured ``container_dict``. Defaults to False.
        cleanup_on_exit : bool
            Determines whether to shut down the connected Fluent session upon exit or when calling
            the session's `exit()` method. Defaults to True.
        start_transcript : bool
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        py : bool, optional
            If True, runs Fluent in Python mode. Defaults to None.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        start_watchdog : bool, optional
            If True and `cleanup_on_exit` is True, an independent watchdog process is run to ensure that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any, optional
            Service for uploading/downloading files to/from the server.

        Returns
        -------
        Meshing | PureMeshing | Solver | SolverIcing | dict
            Session object or configuration dictionary if ``dry_run`` is True.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        mode = cls._session_mode[cls.__name__]
        argvals = locals().copy()
        argvals.pop("cls", None)
        launcher = DockerLauncher(**argvals)
        return launcher()

    @classmethod
    def from_pim(
        cls,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        cleanup_on_exit: bool = True,
        start_transcript: bool = True,
        gpu: bool | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
    ):
        """
        Launch a Fluent session in `PIM <https://pypim.docs.pyansys.com/version/stable/>`_ mode.

        Parameters
        ----------
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version : FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            any of ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values include ``Dimension.TWO``,
            ``Dimension.THREE``, or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values include ``Precision.SINGLE``,
            ``Precision.DOUBLE``, or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        start_timeout : int, optional
            Maximum allowable time in seconds for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        cleanup_on_exit : bool
            Determines whether to shut down the connected Fluent session upon exit or when calling
            the session's `exit()` method. Defaults to True.
        start_transcript : bool
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        start_watchdog : bool, optional
            If True and `cleanup_on_exit` is True, an independent watchdog process is run to ensure that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any, optional
            Service for uploading/downloading files to/from the server.

        Returns
        -------
        Union[Meshing, PureMeshing, Solver, SolverIcing, dict]
            Session object or configuration dictionary if ``dry_run`` is True.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        mode = cls._session_mode[cls.__name__]
        argvals = locals().copy()
        argvals.pop("cls", None)
        launcher = PIMLauncher(**argvals)
        return launcher()

    @classmethod
    def from_connection(
        cls,
        ip: str | None = None,
        port: int | None = None,
        server_info_file_name: str | None = None,
        password: str | None = None,
    ):
        """Connect to an existing Fluent server instance.

        Parameters
        ----------
        ip : str, optional
            IP address for connecting to an existing Fluent instance. The
            IP address defaults to ``"127.0.0.1"``. You can also use the environment
            variable ``PYFLUENT_FLUENT_IP=<ip>`` to set this parameter.
            The explicit value of ``ip`` takes precedence over ``PYFLUENT_FLUENT_IP=<ip>``.
        port : int, optional
            Port to listen on for an existing Fluent instance. You can use the
            environment variable ``PYFLUENT_FLUENT_PORT=<port>`` to set a default
            value. The explicit value of ``port`` takes precedence over
            ``PYFLUENT_FLUENT_PORT=<port>``.
        server_info_file_name: str
            Path to server-info file written out by Fluent server. The default is
            ``None``. PyFluent uses the connection information in the file to
            connect to a running Fluent session.
        password : str, optional
            Password to connect to existing Fluent instance.

        Raises
        ------
        TypeError
            If the session type does not match the expected session type.
        """
        session = pyfluent.connect_to_fluent(
            ip=ip,
            port=port,
            server_info_file_name=server_info_file_name,
            password=password,
        )

        expected = "Solver" if cls.__name__ == "PrePost" else cls.__name__
        actual = session.__class__.__name__

        if actual != expected:
            raise TypeError(
                f"Session type mismatch: expected {expected}, got {actual}."
            )

        return session


class Meshing(SessionBase):
    """Encapsulates a Fluent server for meshing session connection."""

    pass


class PureMeshing(SessionBase):
    """Encapsulates a Fluent server for pure meshing session connection."""

    pass


class PrePost(SessionBase):
    """Encapsulates a Fluent server for pre-post session connection."""

    pass


class Solver(SessionBase):
    """Encapsulates a Fluent server for solver session connection."""

    pass


class SolverAero(SessionBase):
    """Encapsulates a Fluent server for solver aero session connection."""

    pass


class SolverIcing(SessionBase):
    """Encapsulates a Fluent server for solver icing session connection."""

    pass



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Wrappers over Reduction gRPC service of Fluent."""

from typing import Any, List, Tuple
import weakref

import grpc

from ansys.api.fluent.v0 import reduction_pb2 as ReductionProtoModule
from ansys.api.fluent.v0 import reduction_pb2_grpc as ReductionGrpcModule
from ansys.fluent.core.services.datamodel_se import _convert_variant_to_value
from ansys.fluent.core.services.interceptors import (
    BatchInterceptor,
    ErrorStateInterceptor,
    GrpcErrorInterceptor,
    TracingInterceptor,
)
from ansys.fluent.core.variable_strategies import (
    FluentExprNamingStrategy as naming_strategy,
)

Path = List[Tuple[str, str]]


class ReductionService:
    """Reduction Service."""

    def __init__(
        self, channel: grpc.Channel, metadata: List[Tuple[str, str]], fluent_error_state
    ):
        """__init__ method of Reduction class."""
        intercept_channel = grpc.intercept_channel(
            channel,
            GrpcErrorInterceptor(),
            ErrorStateInterceptor(fluent_error_state),
            TracingInterceptor(),
            BatchInterceptor(),
        )
        self._stub = ReductionGrpcModule.ReductionStub(intercept_channel)
        self._metadata = metadata

    def area(
        self, request: ReductionProtoModule.AreaRequest
    ) -> ReductionProtoModule.AreaResponse:
        """Area RPC of Reduction service."""
        return self._stub.Area(request, metadata=self._metadata)

    def area_average(
        self, request: ReductionProtoModule.AreaAveRequest
    ) -> ReductionProtoModule.AreaAveResponse:
        """Area average RPC of Reduction service."""
        return self._stub.AreaAve(request, metadata=self._metadata)

    def area_integral(
        self, request: ReductionProtoModule.AreaIntRequest
    ) -> ReductionProtoModule.AreaIntResponse:
        """Area integral RPC of Reduction service."""
        return self._stub.AreaInt(request, metadata=self._metadata)

    def centroid(
        self, request: ReductionProtoModule.CentroidRequest
    ) -> ReductionProtoModule.CentroidResponse:
        """Centroid RPC of Reduction service."""
        return self._stub.Centroid(request, metadata=self._metadata)

    def count(
        self, request: ReductionProtoModule.CountRequest
    ) -> ReductionProtoModule.CountResponse:
        """Count RPC of Reduction service."""
        return self._stub.Count(request, metadata=self._metadata)

    def count_if(
        self, request: ReductionProtoModule.CountIfRequest
    ) -> ReductionProtoModule.CountIfResponse:
        """Count If RPC of Reduction service."""
        return self._stub.CountIf(request, metadata=self._metadata)

    def force(
        self, request: ReductionProtoModule.ForceRequest
    ) -> ReductionProtoModule.ForceResponse:
        """Force RPC of Reduction service."""
        return self._stub.Force(request, metadata=self._metadata)

    def mass_average(
        self, request: ReductionProtoModule.MassAveRequest
    ) -> ReductionProtoModule.MassAveResponse:
        """Mass average RPC of Reduction service."""
        return self._stub.MassAve(request, metadata=self._metadata)

    def mass_flow_average(
        self, request: ReductionProtoModule.MassFlowAveRequest
    ) -> ReductionProtoModule.MassFlowAveResponse:
        """Mass flow average RPC of Reduction service."""
        return self._stub.MassFlowAve(request, metadata=self._metadata)

    def mass_flow_average_absolute(
        self, request: ReductionProtoModule.MassFlowAveAbsRequest
    ) -> ReductionProtoModule.MassFlowAveAbsResponse:
        """Absolute mass flow average RPC of Reduction service."""
        return self._stub.MassFlowAveAbs(request, metadata=self._metadata)

    def mass_flow_integral(
        self, request: ReductionProtoModule.MassFlowIntRequest
    ) -> ReductionProtoModule.MassFlowIntResponse:
        """Mass flow integral RPC of Reduction service."""
        return self._stub.MassFlowInt(request, metadata=self._metadata)

    def mass_integral(
        self, request: ReductionProtoModule.MassIntRequest
    ) -> ReductionProtoModule.MassIntResponse:
        """Mass integral RPC of Reduction service."""
        return self._stub.MassInt(request, metadata=self._metadata)

    def maximum(
        self, request: ReductionProtoModule.MaximumRequest
    ) -> ReductionProtoModule.MaximumResponse:
        """Maximum RPC of Reduction service."""
        return self._stub.Maximum(request, metadata=self._metadata)

    def minimum(
        self, request: ReductionProtoModule.MinimumRequest
    ) -> ReductionProtoModule.MinimumResponse:
        """Minimum RPC of Reduction service."""
        return self._stub.Minimum(request, metadata=self._metadata)

    def pressure_force(
        self, request: ReductionProtoModule.PressureForceRequest
    ) -> ReductionProtoModule.PressureForceResponse:
        """Pressure force RPC of Reduction service."""
        return self._stub.PressureForce(request, metadata=self._metadata)

    def viscous_force(
        self, request: ReductionProtoModule.ViscousForceRequest
    ) -> ReductionProtoModule.ViscousForceResponse:
        """Viscous force RPC of Reduction service."""
        return self._stub.ViscousForce(request, metadata=self._metadata)

    def volume(
        self, request: ReductionProtoModule.VolumeRequest
    ) -> ReductionProtoModule.VolumeResponse:
        """Volume RPC of Reduction service."""
        return self._stub.Volume(request, metadata=self._metadata)

    def volume_average(
        self, request: ReductionProtoModule.VolumeAveRequest
    ) -> ReductionProtoModule.VolumeAveResponse:
        """Volume average RPC of Reduction service."""
        return self._stub.VolumeAve(request, metadata=self._metadata)

    def volume_integral(
        self, request: ReductionProtoModule.VolumeIntRequest
    ) -> ReductionProtoModule.VolumeIntResponse:
        """Volume integral RPC of Reduction service."""
        return self._stub.VolumeInt(request, metadata=self._metadata)

    def moment(
        self, request: ReductionProtoModule.MomentRequest
    ) -> ReductionProtoModule.MomentResponse:
        """Moment RPC of Reduction service."""
        return self._stub.Moment(request, metadata=self._metadata)

    def sum(
        self, request: ReductionProtoModule.SumRequest
    ) -> ReductionProtoModule.SumResponse:
        """Sum RPC of Reduction service."""
        return self._stub.Sum(request, metadata=self._metadata)

    def sum_if(
        self, request: ReductionProtoModule.SumIfRequest
    ) -> ReductionProtoModule.SumIfResponse:
        """Sum RPC of Reduction service."""
        return self._stub.SumIf(request, metadata=self._metadata)


class BadReductionRequest(Exception):
    """Raised on an attempt to make a bad reduction request."""

    def __init__(self, err):
        """Initialize BadReductionRequest."""
        super().__init__(f"Could not complete reduction function request: {err}")


def _validate_locn_list(locn_list, ctxt):
    if not all(locn[0] for locn in locn_list) and (
        any(locn[0] for locn in locn_list) or not ctxt
    ):
        raise BadReductionRequest("Invalid combination of arguments")


def _is_iterable(obj):
    return hasattr(type(obj), "__iter__")


def _expand_locn_container(locns):
    try:
        return [[locn, locns] for locn in locns]
    except TypeError as ex:
        raise BadReductionRequest(ex)


def _locn_name_and_obj(locn, locns):
    if isinstance(locn, str):
        return [locn, locns]
    # should call locn_get_name()
    if _is_iterable(locn):
        return _locn_names_and_objs(locn)
    else:
        return [locn.obj_name, locn]


def _locn_names_and_objs(locns):
    if _is_iterable(locns):
        names_and_objs = []
        for locn in locns:
            name_and_obj = _locn_name_and_obj(locn, locns)
            if _is_iterable(name_and_obj):
                if isinstance(name_and_obj[0], str):
                    names_and_objs.append(name_and_obj)
                else:
                    names_and_objs.extend(name_and_obj)
        return names_and_objs
    else:
        return _expand_locn_container(locns)


def _root(obj):
    return (
        None
        if isinstance(obj, list)
        else obj if not getattr(obj, "obj_name", None) else _root(obj._parent)
    )


def _locns(locns, ctxt=None):
    locn_names_and_objs = _locn_names_and_objs(locns)
    locn_list = []
    for name, obj in locn_names_and_objs:
        root = _root(obj)
        found = False
        for locn in locn_list:
            if locn[0] is root:
                locn[1].append(name)
                found = True
                break
        if not found:
            locn_list.append([root, [name]])
    _validate_locn_list(locn_list, ctxt)
    return locn_list


class Reduction:
    """Reduction."""

    def __init__(self, service: ReductionService, ctxt=None):
        """__init__ method of Reduction class."""
        self.service = service
        self.ctxt = weakref.proxy(ctxt)
        self._to_str = naming_strategy().to_string

    def _validate_str_location(self, loc: str):
        if all(
            loc not in names()
            for names in (
                self.ctxt.fields.field_info.get_surfaces_info,
                self.ctxt.settings.setup.cell_zone_conditions,
            )
        ):
            raise ValueError(f"Invalid location input: '{loc}'")

    def _get_location_string(self, locations, ctxt) -> List[str]:
        if locations == []:
            return []
        for loc in locations:
            if isinstance(loc, str):
                self._validate_str_location(loc)
        try:
            return _locns(locations, ctxt)[0][1]
        except BadReductionRequest:
            return locations

    def _make_request(
        self,
        requestName,
        locations,
        ctxt=None,
        expression=None,
        weight=None,
        condition=None,
    ) -> Any:
        request = getattr(ReductionProtoModule, requestName)()
        if expression is not None:
            request.expression = self._to_str(expression)
        if weight is not None:
            request.weight = weight
        if condition is not None:
            request.condition = condition
        request.locations.extend(self._get_location_string(locations, ctxt))
        return request

    def area(self, locations, ctxt=None) -> Any:
        """Get area."""
        request = self._make_request("AreaRequest", locations, ctxt)
        response = self.service.area(request)
        return _convert_variant_to_value(response.value)

    def area_average(self, expression, locations, ctxt=None) -> Any:
        """Get area average."""
        request = self._make_request("AreaAveRequest", locations, ctxt, expression)
        response = self.service.area_average(request)
        return _convert_variant_to_value(response.value)

    def area_integral(self, expression, locations, ctxt=None) -> Any:
        """Get area integral."""
        request = self._make_request("AreaIntRequest", locations, ctxt, expression)
        response = self.service.area_integral(request)
        return _convert_variant_to_value(response.value)

    def centroid(self, locations, ctxt=None) -> Any:
        """Get centroid."""
        request = self._make_request("CentroidRequest", locations, ctxt)
        response = self.service.centroid(request)
        return (response.value.x, response.value.y, response.value.z)

    def count(self, locations, ctxt=None) -> Any:
        """Count the number of faces or cells within the locations."""
        request = self._make_request("CountRequest", locations, ctxt)
        response = self.service.count(request)
        return _convert_variant_to_value(response.value)

    def count_if(self, condition, locations, ctxt=None) -> Any:
        """Count the number of faces or cells where the specified condition is satisfied."""
        request = self._make_request(
            "CountIfRequest", locations, ctxt, expression=condition
        )
        response = self.service.count_if(request)
        return _convert_variant_to_value(response.value)

    def force(self, locations, ctxt=None) -> Any:
        """Get force."""
        request = self._make_request("ForceRequest", locations, ctxt)
        response = self.service.force(request)
        return (response.value.x, response.value.y, response.value.z)

    def mass_average(self, expression, locations, ctxt=None) -> Any:
        """Get mass average."""
        request = self._make_request("MassAveRequest", locations, ctxt, expression)
        response = self.service.mass_average(request)
        return _convert_variant_to_value(response.value)

    def mass_flow_average(self, expression, locations, ctxt=None) -> Any:
        """Get mass flow average."""
        request = self._make_request("MassFlowAveRequest", locations, ctxt, expression)
        response = self.service.mass_flow_average(request)
        return _convert_variant_to_value(response.value)

    def mass_flow_average_absolute(self, expression, locations, ctxt=None) -> Any:
        """Compute the mass flow average of the absolute value of the given expression."""
        request = self._make_request(
            "MassFlowAveAbsRequest", locations, ctxt, expression
        )
        response = self.service.mass_flow_average_absolute(request)
        return _convert_variant_to_value(response.value)

    def mass_flow_integral(self, expression, locations, ctxt=None) -> Any:
        """Get mass flow integral."""
        request = self._make_request("MassFlowIntRequest", locations, ctxt, expression)
        response = self.service.mass_flow_integral(request)
        return _convert_variant_to_value(response.value)

    def mass_integral(self, expression, locations, ctxt=None) -> Any:
        """Get mass integral."""
        request = self._make_request("MassIntRequest", locations, ctxt, expression)
        response = self.service.mass_integral(request)
        return _convert_variant_to_value(response.value)

    def maximum(self, expression, locations, ctxt=None) -> Any:
        """Get maximum."""
        request = self._make_request("MaximumRequest", locations, ctxt, expression)
        response = self.service.maximum(request)
        return _convert_variant_to_value(response.value)

    def minimum(self, expression, locations, ctxt=None) -> Any:
        """Get minimum."""
        request = self._make_request("MinimumRequest", locations, ctxt, expression)
        response = self.service.minimum(request)
        return _convert_variant_to_value(response.value)

    def pressure_force(self, locations, ctxt=None) -> Any:
        """Get pressure force."""
        request = self._make_request("PressureForceRequest", locations, ctxt)
        response = self.service.pressure_force(request)
        return (response.value.x, response.value.y, response.value.z)

    def viscous_force(self, locations, ctxt=None) -> Any:
        """Get viscous force."""
        request = self._make_request("ViscousForceRequest", locations, ctxt)
        response = self.service.viscous_force(request)
        return (response.value.x, response.value.y, response.value.z)

    def volume(self, locations, ctxt=None) -> Any:
        """Get volume."""
        request = self._make_request("VolumeRequest", locations, ctxt)
        response = self.service.volume(request)
        return _convert_variant_to_value(response.value)

    def volume_average(self, expression, locations, ctxt=None) -> Any:
        """Get volume average."""
        request = self._make_request("VolumeRequest", locations, ctxt, expression)
        response = self.service.volume_average(request)
        return _convert_variant_to_value(response.value)

    def volume_integral(self, expression, locations, ctxt=None) -> Any:
        """Get volume integral."""
        request = self._make_request("VolumeIntRequest", locations, ctxt, expression)
        response = self.service.volume_integral(request)
        return _convert_variant_to_value(response.value)

    def moment(self, expression, locations, ctxt=None) -> Any:
        """Get moment."""
        request = self._make_request("MomentRequest", locations, ctxt, expression)
        response = self.service.moment(request)
        return (response.value.x, response.value.y, response.value.z)

    def sum(self, expression, locations, weight, ctxt=None) -> Any:
        """Get sum."""
        request = self._make_request("SumRequest", locations, ctxt, expression, weight)
        response = self.service.sum(request)
        return _convert_variant_to_value(response.value)

    def sum_if(self, expression, condition, locations, weight, ctxt=None) -> Any:
        """Compute the weighted sum of the expression at locations where the given condition is satisfied."""
        request = self._make_request(
            "SumIfRequest", locations, ctxt, expression, weight, condition
        )
        response = self.service.sum_if(request)
        return _convert_variant_to_value(response.value)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for customized error handling."""

from ansys.fluent.core.exceptions import InvalidArgument
from ansys.fluent.core.launcher import launcher_utils
from ansys.fluent.core.utils.fluent_version import FluentVersion


class InvalidPassword(ValueError):
    """Raised when password is invalid."""

    def __init__(self):
        """Initialize InvalidPassword."""
        super().__init__("Provide correct 'password'.")


class GPUSolverSupportError(ValueError):
    """Raised when an unsupported Fluent version is specified."""

    def __init__(self):
        """Initialize GPUSolverSupportError."""
        super().__init__("Fluent GPU Solver is only supported for 3D.")


class IpPortNotProvided(ValueError):
    """Raised when IP address and port are not specified."""

    def __init__(self):
        """Initialize IpPortNotProvided."""
        super().__init__("Provide either 'ip' and 'port' or 'server_info_file_name'.")


class InvalidIpPort(ValueError):
    """Raised when IP address and port are invalid."""

    def __init__(self):
        """Initialize InvalidIpPort."""
        super().__init__("Provide a valid 'ip' and 'port'.")


class UnexpectedKeywordArgument(TypeError):
    """Raised when a valid keyword argument is not specified."""

    pass


class LaunchFluentError(Exception):
    """Exception class representing launch errors."""

    def __init__(self, launch_string):
        """__init__ method of LaunchFluentError class."""
        details = "\n" + "Fluent Launch string: " + launch_string
        super().__init__(details)


def _raise_non_gui_exception_in_windows(
    ui_mode, product_version: FluentVersion
) -> None:
    """Fluent user interface mode lower than ``UIMode.HIDDEN_GUI`` is not supported in
    Windows in Fluent versions earlier than 2024 R1."""
    from ansys.fluent.core.launcher.launch_options import UIMode

    if (
        launcher_utils.is_windows()
        and UIMode(ui_mode) not in [UIMode.GUI, UIMode.HIDDEN_GUI]
        and product_version < FluentVersion.v241
    ):
        raise InvalidArgument(
            f"'{ui_mode}' supported in Windows only for {str(FluentVersion.v241)} or later."
        )


def _process_kwargs(kwargs):
    """Verify whether keyword arguments are valid or not.

    Parameters
    ----------
    kwargs: Any
        Keyword arguments.

    Raises
    ------
    UnexpectedKeywordArgument
        If an unexpected keyword argument is provided.
    """
    if kwargs:
        if "meshing_mode" in kwargs:
            raise UnexpectedKeywordArgument(
                "Use 'launch_fluent(mode='meshing')' to launch Fluent in meshing mode."
            )
        else:
            raise UnexpectedKeywordArgument(
                f"launch_fluent() got an unexpected keyword argument {next(iter(kwargs))}"
            )
